/*
 * riproute.h
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Revision: 1.1 $
 * $Date: 1994/09/15 20:56:18 $
 */

/*
 * Interface to the routing subsystem
 */

#ifndef riproute_h
#define riproute_h

#define RIPROUTE_TTL_INFINITE	0xffff
#define RTDEFAULTTTL    60*24	  /* default ttl for a route - 1 day */

typedef struct route {
	IPhost	net; 		/* net for this route */
	IPhost  mask;		/* mask for this net */
 	IPhost  gw;		/* gateway for hop */
	int 	metric; 	/* distance metric */
	Sessn 	netSessn; 	/* cached network session */
	u_short key;		/* sort key */
	u_short ttl;		/* time to live */
	int	ifnum;		/* interface number */
	struct route *next;	/* next for this hash */
} route;


typedef struct {
	u_char  valid;		/* is table valid */
	route	*defrt;		/* default route */
	u_short bpoolsize;	/* number of unallocated structures remaining */
	route	**arr;
} RipRouteTable;


#include "rip_i.h"

#ifdef __STDC__

/* 
 * Initialize the routing tables and set the default router to be the
 * given RIP host.  If the RIP host is all zeroes or if it is not
 * directly reachable, there will be no default router and ript_init
 * will return XK_FAILURE.
 */
xkern_return_t 	ript_init( PSTATE *, IPhost * );
xkern_return_t 	ript_add( PSTATE *, IPhost *, IPhost *, IPhost *, int, int, int );
xkern_return_t	ript_add_def( PSTATE *, IPhost *, int );
xkern_return_t	ript_get( RipRouteTable *, IPhost *, route * );
void		ript_delete( RipRouteTable *, IPhost *, IPhost * );
void     	ript_free( RipRouteTable *, route * );

#else

xkern_return_t 	ript_init();
xkern_return_t 	ript_add();
xkern_return_t	ript_add_def();
xkern_return_t	ript_get();
void		ript_delete();
void     	ript_free();


#endif /* __STDC__ */

#endif /* ! riproute_h */
