#include "ipf.h"
#include "xkernel.h"
#include "eth_i.h"
#include "ip.h"
#include "x_stdio.h"

#define TRACE_VAR ipf_icmpp

typedef struct {
  long* MY_HOST;
  long* CLIENT_HOST;
  ETHhost* CLIENT_ETH;
} PState;

int traceipf_icmpp;

int                     ipf_icmp_init (Protl);
static xkern_return_t   ipf_icmpDemux( Protl, Sessn, Msg * );

int ipf_icmp_init (self)
     Protl self;
{
  Protl llp;
  Part p;
  PState* s;

  printf("\nI am the IP Forwarder for ICMP.\n\n");

  self->demux = ipf_icmpDemux;

  s = (PState*) malloc(sizeof(PState));
  self->state = (char*) s;
  s->MY_HOST = (long*) malloc(sizeof(long));
  s->CLIENT_HOST = (long*) malloc(sizeof(long));
  s->CLIENT_ETH = (ETHhost*) malloc(sizeof(ETHhost));

  str2ipHost((IPhost*) s->MY_HOST, MY_IP);
  str2ipHost((IPhost*) s->CLIENT_HOST, MY_CLIENT_IP);
  MY_CLIENT_ETH(s->CLIENT_ETH);

  llp = xGetDown(self, 0);
  partInit(&p, 1);
  partPush(p, ANY_HOST, 0);
  xOpenEnable(self, self, llp, &p);

  return 0;
}


static xkern_return_t ipf_icmpDemux( self, lls, dg )
    Protl self;
    Sessn lls;
    Msg *dg;
{
  PState* s = (PState *)self->state;
  Sessn lls_new;
  Part p[2];
  Part* tmp_p;
  char xc_buf[1000];
  int p_size;
  long* local_h;
  long* remote_h;
  IPhost tmp_l, tmp_r;
  IPhost tmp_host1;
  IPhost* tmp_host2;
  ETHhdr* eth_hdr;

  xTrace0(TRACE_VAR, TR_EVENTS, "***Got a message!");

  xTrace0(TRACE_VAR, TR_EVENTS, "***Getting participants list from IP");

  xControl(lls, GETPARTICIPANTS, xc_buf, 1000);
  p_size = partExtLen(xc_buf);
  tmp_p = (Part*) malloc(sizeof(Part)*p_size);
  partInternalize(tmp_p, xc_buf);

  remote_h = (long*) partPop(tmp_p[0]);
  local_h = (long*) partPop(tmp_p[1]);

  tmp_l = *((IPhost*) local_h);
  tmp_r = *((IPhost*) remote_h);

  xTrace4(TRACE_VAR, TR_EVENTS, "***Original Local Address: %d.%d.%d.%d", tmp_l.a, tmp_l.b, tmp_l.c, tmp_l.d);
  xTrace4(TRACE_VAR, TR_EVENTS, "***Original Remote Address: %d.%d.%d.%d", tmp_r.a, tmp_r.b, tmp_r.c, tmp_r.d);

  xTrace0(TRACE_VAR, TR_EVENTS, "***Changing participants");

  partInit(p, 2);

  if(*local_h == *(s->MY_HOST)) {
    xTrace0(TRACE_VAR, TR_EVENTS, "***This is a message from the outside world");

    xTrace0(TRACE_VAR, TR_EVENTS, "***Getting an IP Pseudoheader");
    xControl(lls, IP_GETPSEUDOHDR, xc_buf, 1000);
    tmp_host1 = (((IPpseudoHdr *) xc_buf)->src);

    tmp_host2 = (IPhost*) malloc(sizeof(IPhost));
    str2ipHost(tmp_host2, MY_CLIENT_IP);

    if(tmp_host1.a==255) tmp_host2->a = 255;
    if(tmp_host1.b==255) tmp_host2->b = 255;
    if(tmp_host1.c==255) tmp_host2->c = 255;
    if(tmp_host1.d==255) tmp_host2->d = 255;

    partPush(p[0], (void*) tmp_host2, sizeof(IPhost));
    partPush(p[1], (void*) remote_h, sizeof(IPhost));

    tmp_l = *((IPhost*) remote_h);
    tmp_r = *tmp_host2;

  } else if (*remote_h == *(s->CLIENT_HOST)){
    xTrace0(TRACE_VAR, TR_EVENTS, "***This seems to be a message from the IPF Client");
    xTrace0(TRACE_VAR, TR_EVENTS, "***Attempting to verify...");

    eth_hdr = (ETHhdr*) msgGetAttr(dg, 0);

    xTrace3(TRACE_VAR, TR_EVENTS, "***Src eth addr: %u.%u.%u",eth_hdr->src.high,eth_hdr->src.mid,eth_hdr->src.low);
    xTrace3(TRACE_VAR, TR_EVENTS, "***Dst eth addr: %u.%u.%u",eth_hdr->dst.high,eth_hdr->dst.mid,eth_hdr->dst.low);

    if(!(ETH_ADS_EQUAL(eth_hdr->src, (*(s->CLIENT_ETH))))) {
      xTrace0(TRACE_VAR, TR_ERRORS, "***Houston, we have a problem...");
      xTrace0(TRACE_VAR, TR_ERRORS, "***Fake IP address, ETH didn't check out...");
      return XK_FAILURE;
    }

    xTrace0(TRACE_VAR, TR_EVENTS, "***Verified.");
    xTrace0(TRACE_VAR, TR_EVENTS, "***This is indeed a message from the IPF Client");
    xTrace0(TRACE_VAR, TR_EVENTS, "***Getting an IP Pseudoheader");

    xControl(lls, IP_GETPSEUDOHDR, xc_buf, 1000);
    remote_h = (long*) &(((IPpseudoHdr *) xc_buf)->src);
    partPush(p[0], (void*) remote_h, sizeof(IPhost));
    partPush(p[1], (void*) s->MY_HOST, sizeof(IPhost));

    tmp_l = *((IPhost*) s->MY_HOST);
    tmp_r = *((IPhost*) remote_h);
  } else {
    xTrace0(TRACE_VAR, TR_EVENTS, "***This message is not for me. Dropping...");
    return XK_SUCCESS;
  }

  xTrace4(TRACE_VAR, TR_EVENTS, "***New Local Address: %d.%d.%d.%d", tmp_l.a, tmp_l.b, tmp_l.c, tmp_l.d);
  xTrace4(TRACE_VAR, TR_EVENTS, "***New Remote Address: %d.%d.%d.%d", tmp_r.a, tmp_r.b, tmp_r.c, tmp_r.d);

  xTrace0(TRACE_VAR, TR_EVENTS, "***Forwarding the message");

  lls_new = xOpen(self, self, xGetDown(self, 0), p);
  if(lls_new != ERR_XOBJ) {
    if(xPush (lls_new, dg) == XK_FAILURE) return XK_FAILURE;
    if(xClose(lls_new) == XK_FAILURE) return XK_FAILURE;
    return XK_SUCCESS;
  } else {
    xTrace0(TRACE_VAR, TR_ERRORS, "***Houston, we have a problem...");
    xTrace0(TRACE_VAR, TR_ERRORS, "***Can't xOpen an IP session...");
    return XK_FAILURE;
  }
}

