/*     
 * $RCSfile: dispatch.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: dispatch.c,v $
 * Revision 1.2  1994/05/06 20:31:11  menze
 *   [ 1994/05/06          menze ]
 *   Fixed dispatch_open declaration problem
 *   Use msgShow rather than MSG_SHOW
 *
 * Revision 1.1  1994/04/18  22:02:04  slm
 * Initial revision
 *
 */

#include "xkernel.h"
#include "psync.h"
#include "dispatch.h"
#include "dispatch_internal.h"

#define MSG_SHOW	msgShow

       void           dispatch_init(XObj);
static void           getproc_protl(XObj), getproc_sessn(XObj);
static XObj           dispatch_open(XObj, XObj, XObj, Part *);
static xkern_return_t dispatch_demux(XObj, XObj, Msg *);
static int            dispatch_controlprotl(XObj, int, char *, int);
static xmsg_handle_t  dispatch_push(XObj, Msg *);
static xkern_return_t dispatch_pop(XObj, XObj, Msg *, VOID *);
static int            dispatch_controlsessn(XObj, int, char *, int);
static long           dispatch_hdr_peek(VOID *, char *, long, VOID *);
static long           dispatch_hdr_strip(VOID *, char *, long, VOID *);
static void           dispatch_hdr_store(VOID *, char *, long, VOID *);	

int tracedispatchp = 1;

void dispatch_init(self)
XObj self;
{
    DISPATCHstate *pstate;

    xAssert(xIsProtocol(self));
    xTrace0(dispatchp, TR_GROSS_EVENTS, "DISPATCH init");
    if (!xIsProtocol(xGetDown(self,0)))
        Kabort("DISPATCH: could not get lower protocol TOTAL_ORDER");

    getproc_protl(self);

    if ((pstate = (DISPATCHstate *)xMalloc(sizeof(DISPATCHstate))) == NULL)
	Kabort("DISPATCH: malloc failed in dispatch_init");
    self->state = (char *)pstate;
    pstate->map_demux = mapCreate(100, sizeof(Pbindkey));
    pstate->map_d     = mapCreate(100, sizeof(XObj));
}

static void getproc_protl(self)
XObj self;
{
    xAssert(xIsProtocol(self));

    self->open    = dispatch_open;
    self->demux   = dispatch_demux;
    self->control = dispatch_controlprotl;
}

static void getproc_sessn(self)
XObj self;
{
    xAssert(xIsSession(self));

    self->push    = dispatch_push;
    self->pop     = dispatch_pop;
    self->control = dispatch_controlsessn;
}

static XObj dispatch_open(self, hlpRcv, hlpType, p)
XObj self, hlpRcv, hlpType;
Part *p;	/* p = [psync1,opcode] */
{
    XObj     psync_sessn, total_sessn, dispatch_sessn;
    Part     part[1];
    Pbindkey bindkey;
    Sstate   *sstate;
    int      opcode;

    xAssert(xIsProtocol(self));
    xTrace0(dispatchp, TR_FUNCTIONAL_TRACE, "DISPATCH: entering dispatch_open");

    psync_sessn = *(XObj *)partPop(p[0]);
    opcode      = *(int *)partPop(p[1]);
    xTrace2(dispatchp, TR_DETAILED, "DISPATCH: PSYNC sessn = %d, opcode = %d",
	    psync_sessn, opcode);

    /*
     * if there is no TOTAL_ORDER session for this PSYNC session
     *     open a TOTAL_ORDER protocol
     *     bind the the returned session to the PSYNC session
     * create a DISPATCH session
     * bind it into the demux_map
     */
    partInit(part, 1);
    /* `map_d' is <psync_sessn> --> <total_session> */
    if (mapResolve(((DISPATCHstate *)self->state)->map_d, (VOID *)&psync_sessn,
		   (VOID **)&total_sessn) == XK_FAILURE) {
        /* there is no TOTAL_ORDER session open for this PSYNC session */
        partPush(part[0], (VOID *)&psync_sessn, sizeof(XObj));
        /*
         * Open the underlying TOTAL_ORDER session and bind it to the PSYNC
	 * session.
         */
        total_sessn = xOpen(self, self, xGetDown(self,0), part);
        mapBind(((DISPATCHstate *)self->state)->map_d, (VOID *)&psync_sessn,
		(VOID *)total_sessn);
    }

    /*
     * Create a DISPATCH session and set its down session to be the TOTAL_ORDER
     * session.
     */
    dispatch_sessn = xCreateSessn(getproc_sessn, hlpRcv, hlpType, self, 1,
				  &total_sessn);

    /* Bind the DISPATCH session to the <total_sessn,opcode>. */
    bindkey.sessn  = (int)total_sessn;
    bindkey.opcode = opcode;
    mapBind(((DISPATCHstate *)self->state)->map_demux, (VOID *)&bindkey,
	    (VOID *)dispatch_sessn);
    if ((sstate = (Sstate *)xMalloc(sizeof(Sstate))) == NULL) {
	xError("DISPATCH: malloc failed in dispatch_open");
	return ERR_XOBJ;
    }
    sstate->opcode      = opcode;
    sstate->psync_sessn = psync_sessn;
    dispatch_sessn->state = (char *)sstate;
    xTrace0(dispatchp, TR_FUNCTIONAL_TRACE, "DISPATCH: exiting dispatch_open");
    /*
     * After this return, COMPOSER sets my hlp to the the protocol above
     * it, i.e., the user protocol.
     */
    return dispatch_sessn;
}

static xkern_return_t dispatch_demux(self, lls, msg)
XObj self, lls;
Msg  *msg;	/* msg = [opcode,data] */
{
    DISPATCHhdr hdr;
    Pbindkey    bindkey;
    XObj        dispatch_sessn;

    xTrace0(dispatchp, TR_FUNCTIONAL_TRACE,
	    "DISPATCH: entering dispatch_demux");

    /* Peek into the header and find the opcode field. */
    msgPop(msg, dispatch_hdr_peek, (VOID *)&hdr, sizeof(DISPATCHhdr),
	   (VOID *)NULL);
    bindkey.sessn  = (int)lls;
    bindkey.opcode = hdr.opcode;

    xTrace2(dispatchp, TR_FUNCTIONAL_TRACE,
   "DISPATCH: trying to resolve <TOTsessn=%d,opcode=%d> for the DISPATCH sessn",
            bindkey.sessn, bindkey.opcode);
    /* Find the DISPATCH session for this <total_sessn,opcode> combination. */
    /* `map_demux' is <total_sessn,opcode> --> <dispatch_sessn> */
    if (mapResolve(((DISPATCHstate *)self->state)->map_demux, (VOID *)&bindkey,
		   (VOID **)&dispatch_sessn) == XK_FAILURE) {
        xError("DISPATCH: demux: can't find DISPATCH session; dropping msg");
	return XK_FAILURE;
    }
    else {
        xAssert(xIsSession(dispatch_sessn));
        xPop(dispatch_sessn, lls, msg, (VOID *)NULL);
    }

    xTrace0(dispatchp, TR_FUNCTIONAL_TRACE, "DISPATCH: exiting dispatch_demux");
    return XK_SUCCESS;
}

static int dispatch_controlprotl(self, opcode, buf, len)
XObj self;
int  opcode, len;
char *buf;
{
/*  int      old_s, new_s, opcode, size = 100;
 *  char     buffer[100];
 *  XObj     hlp, s, s1;
 *  Sstate   *state;
 *  Pbindkey bindkey;
 */

    xAssert(xIsProtocol(self));

    switch (opcode) {
        case DISPATCH_RESTORE_SESSN:
            /* buf = {old_s,new_s,opcode,hlp} */
/*          old_s = *(int *)buf;
 *          new_s = *(int *)(buf + sizeof(int));
 *          opcode    = *(int *)(buf + 2 * sizeof(int));
 *          hlp   = *(XObj *)(buf + 3 * sizeof(int));
 *          if ((s1 = (XObj)map_resolve(((DISPATCHstate *)self->state)->map_d,
 *                                      (char *)&new_s)) == (XObj)-1) {
 *              *(int *)buffer = (int)self;
 *              *(int *)(buffer + sizeof(int)) = old_s;
 *              *(int *)(buffer + 2 * sizeof(int)) = new_s;
 *              s1 = (XObj)x_controlprotl(self->down[0],
 *                                        DIRECTORY_RESTORE_SESSN,
 *                                        buffer, size);
 *              map_bind(((DISPATCHstate *)self->state)->map_d,
 *                       (char *)&new_s,(int)s1);
 *          }
 *          s = x_createsession(hlp, self, 1);
 *          bindkey.sessn = (int)s1;
 *          bindkey.opcode    = opcode;
 *          map_bind(((DISPATCHstate *)self->state)->map_demux,
 *                   (char *)&bindkey, (int)s);
 *          s->down[0] = s1;
 *          state = (Sstate *)calloc(1, sizeof(Sstate));
 *          state->opcode = opcode;
 *          state->psync_sessn = (XObj)new_s;
 *          s->state = (char *)state;
 *          return (int)s;
 */
	    return 0;
	default:
	    return xControl(xGetDown(self,0), opcode, buf, len);
    }
}

static xmsg_handle_t dispatch_push(self, msg)
XObj self;
Msg  *msg;
{
    DISPATCHhdr   hdr;
    xmsg_handle_t retval;

    xAssert(xIsSession(self));
    xTrace0(dispatchp, TR_FUNCTIONAL_TRACE, "DISPATCH: entering dispatch_push");
    xTrace1(dispatchp, TR_DETAILED, "DISPATCH: received msg of len %d",
	    msgLen(msg));

    /*
     * Add the opcode as the header of the current message and push it down.
     *     hdr is [opcode]
     * So now msg looks like [opcode,data].
     */
    hdr.opcode = ((Sstate *)self->state)->opcode;
    xTrace1(dispatchp, TR_DETAILED,
	    "DISPATCH: pushing hdr.opcode = %d onto msg", hdr.opcode);
    msgPush(msg, dispatch_hdr_store, (VOID *)&hdr, sizeof(DISPATCHhdr),
	    (VOID *)NULL);
    xTrace1(dispatchp, TR_DETAILED,
            "DISPATCH: pushing msg of len %d", msgLen(msg));
    xIfTrace(dispatchp, TR_DETAILED) {
        MSG_SHOW(msg);
    }
    retval = xPush(xGetDown(self,0), msg);

    xTrace0(dispatchp, TR_FUNCTIONAL_TRACE, "DISPATCH: exiting dispatch_push");
    return retval;
}

static xkern_return_t dispatch_pop(self, lls, msg, hdrp)
XObj self, lls;
Msg  *msg;
VOID *hdrp;
{
    DISPATCHhdr    hdr;
    xkern_return_t retval;

    xAssert(xIsSession(self));
    xTrace0(dispatchp, TR_FUNCTIONAL_TRACE, "DISPATCH: entering dispatch_pop");
    xTrace1(dispatchp, TR_DETAILED, "DISPATCH: received msg of len %d",
	    msgLen(msg));

    msgPop(msg, dispatch_hdr_strip, (VOID *)&hdr, sizeof(DISPATCHhdr),
	   (VOID *)NULL);
    xTrace1(dispatchp, TR_DETAILED,
            "DISPATCH: stripped hdr.opcode = %d from msg", hdr.opcode);
    xTrace1(dispatchp, TR_DETAILED,
            "DISPATCH: popping msg of len %d", msgLen(msg));
    xIfTrace(dispatchp, TR_DETAILED) {
        MSG_SHOW(msg);
    }
    retval = xDemux(self, msg);

    xTrace0(dispatchp, TR_FUNCTIONAL_TRACE, "DISPATCH: exiting dispatch_pop");
    return retval;
}

static int dispatch_controlsessn(self, opcode, buf, len)
XObj self;
int  opcode, len;
char *buf;
{
/*  Sstate *state = (Sstate *)self->state; */

    switch (opcode) {
/*      case DISPATCH_ISALIVE:
 *          xTrace4(dispatchp, TR_DETAILED,
 *                  "DISPATCH_ISALIVE recd for %d %d %d %d\n",
 *                  ((PSYNCaddr *)buf)->host.a, ((PSYNCaddr *)buf)->host.b,
 *                  ((PSYNCaddr *)buf)->host.c, ((PSYNCaddr *)buf)->host.d);
 *          return xControlSessn(state->psync_sessn, PSYNC_ISALIVE, buf, len));
 */
        default:
            return 0;
    }
}

/*
 * Peek into the DISPATCH header of a msg.
 */
static long dispatch_hdr_peek(hdr, src, len, arg)
VOID *hdr, *arg;
char *src;
long len;
{
    bcopy(src, (char *)hdr, len);
    return 0;
}

/*
 * Strip off the header of a msg.
 */
static long dispatch_hdr_strip(hdr, src, len, arg)
VOID *hdr, *arg;
char *src;
long len;
{
    bcopy(src, (char *)hdr, len);
    /* write the ntohs calls here */
    return len;
}

/*
 * Store the header in a msg.
 */
static void dispatch_hdr_store(hdr, dst, len, arg)
VOID *hdr, *arg;
char *dst;
long len;
{
/*
 *  ((DISPATCHhdr *)hdr)->opcode = htons(((DISPATCHhdr *)hdr)->opcode);
 */
    bcopy((char *)hdr, dst, len);
}

